from typing import Tuple

import numpy as np
import torch
import torch.nn as nn
from captum.attr import DeepLiftShap

from ..utils import check_shape, norm_and_to_uint8


def compute_deep_shap(
        model: nn.Module,
        preprocessed_image: torch.Tensor,
        label: torch.Tensor,
        baseline: str = "zero",
        need_check_shape: bool = False) -> Tuple[np.ndarray, np.ndarray]:
    if need_check_shape:
        if not check_shape(preprocessed_image.shape, '1,c,n,n'):
            raise ValueError(f'image has invalid shape: {preprocessed_image.shape}')

    if baseline == "zero":
        base_distribution = preprocessed_image.new_zeros((10,) + preprocessed_image.shape[1:])
    else:
        raise NotImplementedError
    saliency = DeepLiftShap(model).attribute(preprocessed_image, target=label, baselines=base_distribution)
    # saliency shape: (1, height, width)
    saliency = saliency.detach().cpu().numpy().squeeze(axis=0)
    # saliency shape: (height, width)
    saliency = saliency.mean(0)
    normed_saliency = norm_and_to_uint8(saliency, norm='abs_zero_max')

    if need_check_shape:
        if not check_shape(saliency.shape, 'n,n'):
            raise ValueError(f'saliency has invalid shape: {saliency.shape}')
    return saliency.astype(np.float16), normed_saliency
