from typing import Tuple

import numpy as np
import torch
import torch.nn as nn
from captum.attr import GuidedBackprop

from ..utils import check_shape, norm_and_to_uint8


def compute_guided_backprop(
        model: nn.Module,
        preprocessed_image: torch.Tensor,
        label: torch.Tensor,
        remove_negative: bool = False,
        need_check_shape: bool = False) -> Tuple[np.ndarray, np.ndarray]:
    if need_check_shape:
        if not check_shape(preprocessed_image.shape, '1,c,n,n'):
            raise ValueError(f'image has invalid shape: {preprocessed_image.shape}')

    saliency = GuidedBackprop(model).attribute(preprocessed_image, label)
    # grad shape: (1, height, width)
    grad = saliency.detach().cpu().numpy().squeeze(0)
    # grad shape: (height, width)
    grad = grad.mean(0)

    if remove_negative:
        grad[grad < 0] = 0
    normed_grad = norm_and_to_uint8(grad, norm='abs_zero_max')

    if need_check_shape:
        if not check_shape(grad.shape, 'n,n'):
            raise ValueError(f'grad has invalid shape: {grad.shape}')
    return grad.astype(np.float16), normed_grad
