from typing import Tuple

import numpy as np
import torch
import torch.nn as nn
from captum.attr import LRP

from ..utils import check_shape, norm_and_to_uint8


def compute_lrp(
        model: nn.Module,
        preprocessed_image: torch.Tensor,
        label: torch.Tensor,
        use_label: bool,
        need_check_shape: bool = False) -> Tuple[np.ndarray, np.ndarray]:
    if need_check_shape:
        if not check_shape(preprocessed_image.shape, '1,c,n,n'):
            raise ValueError(f'image has invalid shape: {preprocessed_image.shape}')
    # why not using label: according to the docs of Captum,
    # if the network returns a scalar value per example, no target index is necessary
    saliency = LRP(model).attribute(preprocessed_image, target=label if use_label else None)

    # grad shape: (1, channel, height, width)
    grad = saliency.detach().cpu().numpy().squeeze(axis=0)
    # grad shape: (height, width)
    grad = grad.mean(0)
    normed_grad = norm_and_to_uint8(grad, norm='abs_zero_max')

    if need_check_shape:
        if not check_shape(grad.shape, 'n,n'):
            raise ValueError(f'grad has invalid shape: {grad.shape}')
    return grad.astype(np.float16), normed_grad
