from typing import Dict, Tuple, Union

import yaml

DELETE_KEY = '_delete_'


def read_config(file_path: str) -> Dict:
    with open(file_path, 'rb') as f:
        cfg = yaml.safe_load(f)
    return cfg


def _merge_a_into_b(a: Dict, b: Dict, allow_list_keys: bool = False) -> Dict:
    """Copied from mmengine:
    https://github.com/open-mmlab/mmengine/blob/b1b1f53db2712904c04ab61356fe257eb8e9803a/mmengine/config/config.py#L706 # noqa

    merge dict ``a`` into dict ``b`` (non-inplace).
    Values in ``a`` will overwrite ``b``. ``b`` is copied first to avoid
    in-place modifications.
    Args:
        a: The source dict to be merged into ``b``.
        b: The origin dict to be fetch keys from ``a``.
        allow_list_keys (bool): If True, int string keys (e.g. '0', '1')
          are allowed in source ``a`` and will replace the element of the
          corresponding index in b if b is a list. Defaults to False.
    Returns:
        dict: The modified dict of ``b`` using ``a``.

    """
    b = b.copy()
    for k, v in a.items():
        if allow_list_keys and k.isdigit() and isinstance(b, list):
            k = int(k)
            if len(b) <= k:
                raise KeyError(f'Index {k} exceeds the length of list {b}')
            b[k] = _merge_a_into_b(v, b[k], allow_list_keys)
        elif isinstance(v, dict):
            if k in b and not v.pop(DELETE_KEY, False):
                allowed_types: Union[Tuple, type] = (dict, list) if allow_list_keys else dict
                if not isinstance(b[k], allowed_types):
                    raise TypeError(
                        f'{k}={v} in child config cannot inherit from '
                        f'base because {k} is a dict in the child config '
                        f'but is of type {type(b[k])} in base config. '
                        f'You may set `{DELETE_KEY}=True` to ignore the '
                        f'base config.')
                b[k] = _merge_a_into_b(v, b[k], allow_list_keys)
            else:
                b[k] = dict(v)
        else:
            b[k] = v
    return b


def merge_from_options(cfg: Dict, options: Dict, allow_list_keys: bool = True) -> Dict:
    """Merge options into cfg.

    Copied from mmengine:
    https://github.com/open-mmlab/mmengine/blob/b1b1f53db2712904c04ab61356fe257eb8e9803a/mmengine/config/config.py#L967     # noqa

    The options keys can be chained with ".", see the example below.

    Args:
        cfg: Current config dict.
        options: Options to be merged.
        allow_list_keys: If True, int string keys (e.g. '0', '1') are allowed in
            ``options`` and will replace the element of the corresponding index in the
            config if the config is a list. Defaults to True.

    Examples:
        .. code-block:: python3

           cfg = {'outer': {'inner': 1}}
           options = {'outer.inner': 2}

           cfg_new = merge_from_options(cfg, options)
           assert cfg_new['outer']['inner'] == 2

    """
    option_cfg_dict: dict = {}
    for full_key, v in options.items():
        d = option_cfg_dict
        key_list = full_key.split('.')
        for subkey in key_list[:-1]:
            d.setdefault(subkey, dict())
            d = d[subkey]
        subkey = key_list[-1]
        d[subkey] = v

    return _merge_a_into_b(option_cfg_dict, cfg, allow_list_keys=allow_list_keys)
